import re
import os

from hashlib import sha256

sha256_pattern: re.Pattern = re.compile('[A-Fa-f0-9]{64}')  # non case sensitive matching, can return non unique
hashutils_raise_exception: bool = False


def sha256_from_file(filepath: str) -> list:
    """extract all sha256 values from a file using regex

    filepath {str} -- path to file to search

    return {list} -- list of sha256 values found
    """
    sha256_list: list = []
    if os.path.exists(filepath) and os.path.isfile(filepath):
        with open(filepath, 'r') as fin:
            sha256_list.extend(set(sha256_pattern.findall(fin.read())))
    else:
        if hashutils_raise_exception:
            raise FileNotFoundError(f'Filepath was not found {filepath!r}')
    return sha256_list


def sha256_from_dir(directory: str) -> list:
    """extract all sha256 values from all files inside of a directory using regex

    directory {str} -- path to directory to search

    return {list} -- list of sha256 values found
    """
    sha256_list: list = []
    if os.path.exists(directory) and os.path.isfile(directory):
        files = [
            os.path.join(directory, path) for path in os.listdir(directory)
            if os.path.isfile(os.path.join(directory, path))
        ]
        for filepath in files:
            sha256_list.extend(sha256_from_file(filepath))
    return sha256_list


def sha256_from_dir_map(directory: str) -> dict:
    """extract all sha256 values from all files inside of a directory using regex and map them to their filepath

    directory {str} -- path to directory to search

    return {dict} -- dictionary of filepath as key and found sha256 list as value
    """
    sha256_map: dict = {}
    if os.path.exists(directory) and os.path.isfile(directory):
        files = [
            os.path.join(directory, path) for path in os.listdir(directory)
            if os.path.isfile(os.path.join(directory, path))
        ]
        for filepath in files:
            sha256_map.update({filepath: sha256_from_file(filepath)})
    return sha256_map


def calc_sha256_from_dir(directory: str) -> list:
    """calculate sha256 values from files found in a directory

    directory {str} -- path to directory to search

    return {list} -- list of sha256 values found
    """
    sha256_list: list = []
    if os.path.exists(directory) and os.path.isfile(directory):
        files = [
            os.path.join(directory, path) for path in os.listdir(directory)
            if os.path.isfile(os.path.join(directory, path))
        ]
        for filepath in files:
            with open(filepath, 'rb') as fin:
                sha256_list.append(sha256(fin.read()).hexdigest())
    return sha256_list

def calc_sha256_from_dir_map(directory: str) -> dict:
    """calculate sha256 values from files found in a directory and map them to their filepath

    directory {str} -- path to directory to search

    return {dict} -- dictionary with filepath as key and sha256 as value
    """
    sha256_map: dict = {}
    if os.path.exists(directory) and os.path.isfile(directory):
        files = [
            os.path.join(directory, path) for path in os.listdir(directory)
            if os.path.isfile(os.path.join(directory, path))
        ]
        for filepath in files:
            with open(filepath, 'rb') as fin:
                sha256_map.update({filepath: sha256(fin.read()).hexdigest()})
    return sha256_map


import logging

from functools import partial
from requests import Session, Response
from requests.auth import HTTPBasicAuth
from requests.adapters import HTTPAdapter
from email.mime.application import MIMEApplication
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formatdate
from os.path import basename
from smtplib import SMTP
from typing import Union, Text

class BaseSession(Session):
    retries: int
    verbose: bool

    def __init__(self, max_retries: int = 3, pool_connections: int = 16, pool_maxsize: int = 16,
                 resolve_status_codes: list = None, verbose: bool = False, auth: tuple = None):
        super().__init__()
        adapters = HTTPAdapter(pool_connections=pool_connections, pool_maxsize=pool_maxsize,
                                    max_retries=max_retries)
        self.mount("https://", adapters)
        self.mount('http://', adapters)
        self.retries = max_retries
        self.verbose = verbose
        self.resolve_status_codes = [200, 201, 202, 203, 204, 205, 206, 207, 208, 226]
        if isinstance(resolve_status_codes, int):
            self.resolve_status_codes.append(resolve_status_codes)
        elif isinstance(resolve_status_codes, list):
            for sc in resolve_status_codes:
                if isinstance(sc, int):
                    self.resolve_status_codes.append(sc)

        if auth:
            self.auth = HTTPBasicAuth(*auth)

        logging.basicConfig(
            level=logging.INFO,
            format='[%(asctime)s] %(levelname)s - %(message)s',
            datefmt='%H:%M:%S'
        )
        self.session_logger = logging.getLogger(name='BaseSession')
        self._log_msg_fmt = '{method}, {scheme}, {host}, {path}, {content_size}, {user_agent}, {status_code}'


    def _log_response(self, response: Response) -> None:
        """log each requests/response from resolver"""
        self.session_logger.info(self._log_msg_fmt.format(
            scheme=response.url.split("://")[0], host=response.url.split('/')[2], method=response.request.method,
            path=response.request.path_url, status_code=response.status_code, content_size=len(response.content),
            user_agent=response.request.headers.get("User-Agent", "Unknown")))
        if response.status_code >= 300 and self.verbose:
            self.session_logger.error(f'REPONSE: {response.text}')

    def _resolver(self, request: partial) -> Response:
        """attempt to resolve a requests with an invalid status code

        if the status code of the requests is not one to resolve:
            Default:  [200, 201, 202, 203, 204, 205, 206, 207, 208, 226]
        the requests will be sent up to max_retries or until receiving an accepted status_code

        :param request: partial requests function to be used to attempt and resolve a valid response

        :return: response from the requests
        """
        attempt = 1
        resp = request()
        while attempt <= self.retries and resp.status_code not in self.resolve_status_codes:
            resp = request()
            attempt += 1
        self._log_response(resp)
        return resp

    def get(self, url: Union[Text, bytes], **kwargs) -> Response:
        """ override of requests.Session get
            calls Session.get with resolver

        :param url: url for requests
        :param kwargs: named arguments for requests

        :return: response from requests
        """
        return self._resolver(partial(super(BaseSession, self).get, url, **kwargs))

    def head(self, url: Union[Text, bytes], **kwargs) -> Response:
        """ override of requests.Session head
            calls Session.get with resolver

        :param url: url for requests
        :param kwargs: named arguments for requests

        :return: response from requests
        """
        return self._resolver(partial(super(BaseSession, self).head, url, **kwargs))

    def delete(self, url: Union[Text, bytes], **kwargs) -> Response:
        """ override of requests.Session delete
            calls Session.get with resolver

        :param url: url for requests
        :param kwargs: named arguments for requests

        :return: response from requests
        """
        return self._resolver(partial(super(BaseSession, self).delete, url, **kwargs))

    def patch(self, url: Union[Text, bytes], **kwargs) -> Response:
        """ override of requests.Session patch
            calls Session.get with resolver

        :param url: url for requests
        :param kwargs: named arguments for requests

        :return: response from requests
        """
        return self._resolver(partial(super(BaseSession, self).patch, url, **kwargs))

    def post(self, url: Union[Text, bytes], **kwargs) -> Response:
        """ override of requests.Session post
            calls Session.get with resolver

        :param url: url for requests
        :param kwargs: named arguments for requests

        :return: response from requests
        """
        return self._resolver(partial(super(BaseSession, self).post, url, **kwargs))

    def put(self, url: Union[Text, bytes], **kwargs) -> Response:
        """ override of requests.Session put
            calls Session.get with resolver

        :param url: url for requests
        :param kwargs: named arguments for requests

        :return: response from requests
        """
        return self._resolver(partial(super(BaseSession, self).put, url, **kwargs))


class EmailSession(object):

    def __init__(self, smtp_server, smtp_port, from_address: str = None,
                 authentication_required: bool = False, auth_username: str = None, auth_password: str = None):
        self.host = smtp_server
        self.port = smtp_port
        self.from_address = from_address
        self.smtp_session = SMTP(host=self.host, port=self.port)
        self.smtp_session.starttls()
        if authentication_required:
            if not auth_username:
                raise ValueError(f'{"auth_username"!r} cannot be NoneType if {"authentication_required"!r} is True')
            if not auth_password:
                raise ValueError(f'{"auth_password"!r} cannot be NoneType if {"authentication_required"!r} is True')
            self.smtp_session.login(auth_username, auth_password)

        logging.basicConfig(
            level=logging.INFO,
            format='[%(asctime)s] %(levelname)s - %(message)s',
            datefmt='%H:%M:%S'
        )
        self.email_logger = logging.getLogger(name='Mailer')
        self.log_msg_fmt = 'From: {from_address}, To: {to_addresses}, CC: {cc_addresses}, {subject}, ' \
                           'attachments: {attach_len}'

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.smtp_session.quit()

    def send(self, subject: str, body: str, to_address_list: list, cc_address_list: list = None,
             from_address: str = None, encoding: str = 'html', logo_images: list = None,
             file_attachments: list = None) -> dict:
        """

        :param subject: Subject string for email, required
        :param body: Message content for email, required
        :param to_address_list: addresses to send email to, required
        :param cc_address_list: addresses to cc on email, default: None
        :param from_address: address to send email from, default: None, will use self.from_address if one was given
        :param encoding: encoding for body, default: html
        :param logo_images: list of paths to images to use for logos, default: None
        :param file_attachments: list of paths to attachments for email, default: None

        :return: dict
        """
        assert isinstance(to_address_list, list)
        email_from = from_address or self.from_address
        if not email_from:
            raise ValueError(f'{"email_from"!r} cannot be NoneType if {"from_address"!r} is not set.')

        email = MIMEMultipart()
        email['From'] = email_from
        email['To'] = ', '.join(to_address_list)
        email['Date'] = formatdate(localtime=True)
        if cc_address_list:
            assert isinstance(cc_address_list, list)
            email['Cc'] = ', '.join(cc_address_list)

        email['Subject'] = subject
        email.attach(MIMEText(body, encoding, 'utf-8'))

        if logo_images:
            for lpath in logo_images:
                with open(lpath, 'rb') as fin:
                    email.attach(MIMEImage(fin.read()))
        if file_attachments:
            for fpath in file_attachments:
                filename = basename(fpath)
                with open(fpath, 'rb') as fin:
                    attachment = MIMEApplication(fin.read(), Name=filename)
                    attachment['Content-Disposition'] = f'attachment; filename={filename}'
                    email.attach(attachment)

        full_address_list = to_address_list
        if cc_address_list:
            full_address_list.extend(cc_address_list)
        self.email_logger.info(self.log_msg_fmt.format(from_address=email_from,
                                                       to_addresses=', '.join(to_address_list),
                                                       cc_addresses=', '.join(cc_address_list)
                                                       if cc_address_list else 'None', subject=subject,
                                                       attach_len=len(file_attachments)
                                                       if file_attachments else 0
            )
        )
        return self.smtp_session.sendmail(email_from, full_address_list, email.as_string())





import asyncio
import sys

from functools import update_wrapper
from copy import copy
from concurrent.futures import ThreadPoolExecutor
from typing import Callable, List, NewType
from functools import partial
from colored import fg, style
from tqdm import tqdm

Eventloop = NewType('Eventloop', asyncio.windows_events._WindowsSelectorEventLoop) \
        if sys.platform == 'win32' else NewType('Eventloop', asyncio.unix_events._UnixSelectorEventLoop)


def aioloop(function: Callable, args_list: List[List], loop: Eventloop = None,
                max_async_pool: int = 16, max_futures: int = 100000, disable_progress_bar: bool = False,
                progress_bar_color: str = 'green_3a', progress_bar_format: str= None) -> list:
    """create new aioloop, run, and return results

    :param fn {Callable}: function to map to arguments
    :param args_list {List[List]}: list of arguments to send to function
    :param loop {Eventloop}: a pre-defined asyncio loop
    :param max_async_pool {int}: max async pool, this will define the number of processes to run at once
    :param max_futures {int}: max futures, this will define the number of processes to setup and execute at once.
        If there is a lot of arguments and futures is very large, can cause memory issues.
    :param disable_progress_bar {bool}: disable progress bar from printing
    :param progress_bar_color {str}: color of progress bar; default: green
    :param progress_bar_format {str}: format for progress bar output; default: None


    :return list of results
    """
    async def aioexecutor() -> list:
        """aioexecutor will create futures from args and collect results as they are finished in the async loop

        :return: list -- list of results from aio loop
        """
        results = []
        with ThreadPoolExecutor(max_workers=max_async_pool) as executor:
            for index in range(0, len(args_list), max_futures):
                futures = []
                for fnargs in args_list[index:index+max_futures]:
                    fnkwargs = {}
                    for index in range(0, len(fnargs)):
                        if isinstance(fnargs[index], dict):
                            fnkwargs.update(fnargs.pop(index))
                    futures.append(loop.run_in_executor(executor, partial(function, *fnargs, **fnkwargs)))
                results.extend([
                    await result for result in tqdm(asyncio.as_completed(futures), total=len(futures),
                                                    disable=disable_progress_bar, bar_format=progress_bar_format)
                ])
        return results

    if progress_bar_format is None:
        progress_bar_format = '{l_bar}%s{bar}%s| {n_fmt}/{total_fmt} [{elapsed}<{remaining},' \
                 ' {rate_fmt}{postfix}]' % (fg(progress_bar_color), style.RESET)
    if loop is None:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    return loop.run_until_complete(aioexecutor())


class aiobulk(object):

    def __init__(self, function: Callable):
        self.__self__ = None
        self.__bound__ = function
        update_wrapper(self, function)

    def __call__(self, *args: tuple, **kwargs: dict):
        """call base function
        """
        if self.__self__ is not None:
            args = (self.__self__,) + args
        return self.__bound__(*args, **kwargs)

    def __get__(self, instance, _):
        """update self if instance is found

        :param instance: cls instance
        :return: return self with instance
        """
        if instance is None:
            return self
        else:
            self.__self__ = instance
            return self

    def bulk(self, args_list: List[list], loop: Eventloop = None, max_async_pool: int = 16, max_futures: int = 100000,
             disable_progress_bar: bool = False, progress_bar_color: str = 'green_3a',
             progress_bar_format: str = None) -> list:
        """add a method called 'bulk' to given function

            :param function {Callable}: function to map to arguments
            :param loop {Eventloop}: a pre-defined asyncio loop
            :param max_async_pool {int}: max async pool, this will define the number of processes to run at once
            :param max_futures {int}: max futures, this will define the number of processes to setup and execute at once.
                If there is a lot of arguments and futures is very large, can cause memory issues.
            :param disable_progress_bar {bool}: disable progress bar from printing
            :param progress_bar_color {str}: color of progress bar; default: green
            :param progress_bar_format {str}: format for progress bar output; default: None

            Examples:
                @aiobulk
                def add(x:int,y:int)->int:return x+y


                # call the original function
                add(1,2)

                # call the newly added bulk function
                args = [[x,y] for x in range(0,5) for y in range(5,10)]
                list_of_returns = add.bulk(args)


                @aiobulk
                def get_url(url, params:dict=None):
                    return requests.get(url, params=params)

                args = [
                    ['https://www.google.com', {'params':{'q':'Why is the sky blue?'}}]
                ]
                responses = get_url.bulk(args)


            :return list of results
        """
        if self.__self__ is not None:
            args_list = [[self.__self__] + args for args in args_list]
        return aioloop(self.__bound__, args_list, loop=loop, max_async_pool=max_async_pool, max_futures=max_futures,
                       disable_progress_bar=disable_progress_bar, progress_bar_color=progress_bar_color,
                       progress_bar_format=progress_bar_format)




from time import sleep
from json import dumps
from json.decoder import JSONDecodeError
from typing import Any
from requests import Response
from colored import fg, style

def response_to_str(response: Response) -> str:
    """
    response_to_str - convert Response object to string and interpret status codes
    :param response: Response object to convert to string
    :return: string of Response object
    """
    if response.status_code in [200, 201, 202, 203, 204, 205, 206, 207, 208, 226]:
        try:
            return dumps(response.json(), indent=4)
        except JSONDecodeError:
            return response.text
    elif response.status_code in [401, 403]:
        return f'({response.status_code}) Invalid permissions for requests: {response.url}'
    elif response.status_code == 404:
        return f'({response.status_code}) Unable to find requested resource for: {response.url}'
    elif response.status_code == 400:
        return f'({response.status_code}) Bad request to: {response.url}'
    elif response.status_code >= 500:
        return f'({response.status_code}) Server error - {response.url} - {response.status_code} - {response.text}'
    elif response.status_code == None:
        return f'({response.status_code}) Server error - {response.url} - {response.text}'
    return f'({response.status_code}) {response.text}'


def echo(content: Any, list_delimiter: str = '\n', indent: int = 4, color: str = None, flush: bool = False) -> None:
    if isinstance(content, (tuple, list)):
        content = list_delimiter.join(map(str, content))
    elif isinstance(content, dict):
        content = dumps(content, indent=indent)
    elif isinstance(content, Response):
        content = response_to_str(content)

    if not isinstance(content, str):
        content = str(content)

    if color:
        content = f'{fg(color)}{content}{style.RESET}'

    if flush:
        print(f'{content}', end='\r', flush=flush)
    else:
        print(content)

def scroll(content: list, sleep_timer=1, list_delimiter: str = '\n', indent: int = 4, color: str = None) -> None:
    for item in content:
        echo(item, flush=True, list_delimiter=list_delimiter, indent=indent, color=color)
        sleep(sleep_timer)



from subprocess import Popen, PIPE
from typing import Union, Tuple

def install_package(name: str, force: bool = False, extra_index: str = None,
                    trusted_host: str = None) -> Tuple[str, str]:
    """install a pip3 package using a subprocess in current environment
    
    name {str} -- name of package to install
    force {bool} -- force newest edition
    extra_index {str} -- extra url to index in package manager
    trusted_host {str} -- extra url where package is hosted

    return {tuple} -- return the output, and error of subprocess run
    """
    cmd = ['pip3', 'install', name]
    if extra_index:
        cmd.extend(['--extra-index-url', extra_index])
    if trusted_host:
        cmd.extend(['--trusted-host', trusted_host])
    if force:
        cmd.append('--upgrade')

    out, err = Popen(cmd, stdout=PIPE, stderr=PIPE).communicate()
    return out.decode('utf-8'), err.decode('utf-8')

def update_package(name: str, extra_index: str = None, trusted_host: str = None) -> Tuple[str, str]:
    """update a pip3 package by name, leverages install_package with force = True

    name {str} -- name of package to install
    extra_index {str} -- extra url to index in package manager
    trusted_host {str} -- extra url where package is hosted

    return {tuple} -- return the output, and error of subprocess run
    """
    return install_package(name, force=True, extra_index=extra_index, trusted_host=trusted_host)

def list_packages() -> list:
    """list current pip3 packages in environment

    return {list} -- a list of available packages
    """
    cmd = ['pip3', 'freeze']
    out, _ = Popen(cmd, stdout=PIPE, stderr=PIPE).communicate()
    return [pkg for pkg in out.decode('utf-8').replace('\r','').split('\n') if pkg]

def has_package(name: str, version: Union[str, int, float] = None) -> bool:
    """check if current environment has a package

    name {str} -- name of package to check
    version {Union[str, int, float]} -- OPTIONAL, will append version for a specific version check of package

    return {bool} -- True if package was found and False if not
    """
    pkg = f'{name}=={version}' if version else name
    return any(pkg in pp for pp in list_packages())




from typing import Union

def globpath(filepath: str) -> str:
    """convert a filepath to a glob path, ['/'|'\'] to '.'

    filepath {str} -- filepath to glob

    return {str} -- new globpath
    """
    glob = filepath.replace('/', '.').replace('\\', '.')
    while glob[0] == '.':
        glob = glob[1:]
    return glob

def import_from(globpath: str, name: str) -> Union[object, None]:
    """Import module and return instance of given function name

    globpath {str} -- the filepath in glob format
    name {str} -- the method name to import

    return {Union[object, None]} -- method attribute of import
    """
    module = __import__(globpath, fromlist=[name])
    return getattr(module, name) if hasattr(module, name) else None


from typing import Any, Union
from modutils.hashutils import sha256_pattern

class sha256:
    name = 'sha256'

    def __init__(self, value):
        value_str = str(value)
        self.__value__ = value_str if sha256_pattern.match(value_str) else None
        if not self.__value__:
            raise TypeError(f'{value!r} is not a valid sha256 type')

    def __str__(self):
        return self.__value__.lower()

    def __eq__(self, value):
        if self.__value__.lower() == sha256(value).__value__.lower():
            return True
        return False

def nget(d:dict, *args:Union[str, list]) -> Any:
    """nget - nested get call to easily retrieve nested information with a single call and set a default
    Ex.
        nget(dict, ['key1', 'key2', ..], default)
        nget(dict, key1, key2, .., default)

        nget use an iterable of keys to retrieve nested information and can set a default if a key is not found
    """

    if len(args) == 0:
        raise ValueError(f'At least 1 key is required. None given')

    keys = args[0] if (isinstance(args[0], list) or len(args) == 1) else args[:-1]
    default = args[-1] if len(args) > 1 else None

    if isinstance(d, dict):
        if isinstance(keys, str):
            keys = [keys]
        if isinstance(keys, (tuple, list)):
            for k in keys:
                d = d.get(k, default)
                if d == default:
                    return default
            if d:
                return d
        else:
            raise TypeError(f'Invalid type of iterable for keys: {type(keys)!r}. Must be tuple or list')
    else:
        raise TypeError(f'First argument must be of type dict, given was type: {type(d)!r}')

    return default
